@@ -2,7 +2,8 @@
 
 #
 # dbfilecat.pm
-# Copyright (C) 2013-2018 by John Heidemann <johnh@isi.edu>
+# Copyright (C) 2013 by John Heidemann <johnh@isi.edu>
+# $Id$
 #
 # This program is distributed under terms of the GNU general
 # public license, version 2.  See the file COPYING
@@ -13,7 +14,7 @@ package Fsdb::Filter::dbfilecat;
 
 =head1 NAME
 
-dbfilecat - concatenate two files with identical schema
+dbfilecat - concatinate two files with identical schema
 
 =head1 SYNOPSIS
 
@@ -26,7 +27,7 @@ or
 
 =head1 DESCRIPTION
 
-Concatenate all provided input files,
+Concatinate all provided input files,
 producing one result.
 We remove extra header lines.
 
@@ -143,8 +144,8 @@ L<Fsdb(3)>
 
 use 5.010;
 use strict;
+use Carp qw(croak);
 use Pod::Usage;
-use Carp qw(croak carp);
 
 use Fsdb::Filter;
 use Fsdb::IO::Reader;
@@ -155,7 +156,7 @@ use Fsdb::IO::Writer;
 
     $filter = new Fsdb::Filter::dbmerge(@arguments);
 
-Create a new object, taking command-line arguments.
+Create a new object, taking command-line arugments.
 
 =cut
 
@@ -229,23 +230,20 @@ sub setup($) {
 	push(@{$self->{_inputs}}, '-');
     };
     if ($self->{_xargs} && $#{$self->{_inputs}} > 0) {
-	croak($self->{_prog} . ": --xargs and multiple inputs (perhaps you meant NOT --xargs?).\n");
-    };
-    if (!$self->{_xargs} && $self->{_remove_inputs}) {
-	croak($self->{_prog} . ": --remove_inputs only works with --xargs.\n");
+	croak $self->{_prog} . ": --xargs and multiple inputs (perhaps you meant NOT --xargs?).\n";
     };
 
     if ($self->{_xargs}) {
-	croak($self->{_prog} . ": --xargs and internal error, no input.\n")
+	croak $self->{_prog} . ": --xargs and internal error, no input.\n"
 	    if ($#{$self->{_inputs}} != 0);
 	# have to delay comments in next line because otherwise _out is not yet open
         $self->finish_io_option('inputs', -header => '#fsdb filename', -comment_handler => $self->create_delay_comments_sub);
-	croak($self->{_prog} . ": xargs setup input stream failed " . $self->{_ins}[0]->error() . "\n")
+	croak $self->{_prog} . ": xargs setup input stream failed " . $self->{_ins}[0]->error() . "\n"
 	    if ($self->{_ins}[0]->error());
     } else {
 	$self->finish_io_option('inputs', -comment_handler => $self->create_pass_comments_sub);
 	foreach (@{$self->{_ins}}) {
-	    croak($self->{_prog} . ": input streams have different schemas; cannot concatenate\n")
+	    croak $self->{_prog} . ": input streams have different schemas; cannot concatinate\n"
 		if ($self->{_ins}[0]->compare($_) ne 'identical');
 	};
 	$self->finish_io_option('output', -clone => $self->{_ins}[0]);
@@ -291,16 +289,12 @@ sub run($) {
 	    if (!$self->{_out}) {
 	    	$self->finish_io_option('output', -clone => $this_in);
 	    } else {
-		croak($self->{_prog} . ": input streams have different schemas; cannot concatenate\n")
+		croak $self->{_prog} . ": input streams have different schemas; cannot concatinate\n"
 		    if ($self->{_out}->compare($this_in) ne 'identical');
 	    };
 	    $self->_run_one($this_in);
-	    if ($self->{_remove_inputs}) {
-		unlink($fref->[0]) or
-		    carp $self->{_prog} . ": --remove-inputs, but cannot remove " . $fref->[0] . "\n";
-	    };
 	};
-	croak($self->{_prog} . ": no input with --xargs\n")
+	croak $self->{_prog} . ": no input with --xargs\n"
 	    if ($inputs == 0);
     } else {
 	foreach my $in (@{$self->{_ins}}) {
@@ -314,7 +308,7 @@ sub run($) {
 
 =head1 AUTHOR and COPYRIGHT
 
-Copyright (C) 2013-2018 by John Heidemann <johnh@isi.edu>
+Copyright (C) 2013 by John Heidemann <johnh@isi.edu>
 
 This program is distributed under terms of the GNU general
 public license, version 2.  See the file COPYING
